import os
import grp


async def load_group(hub):
    hub.grains.GRAINS.gid = os.getegid()
    try:
        hub.grains.GRAINS.groupname = grp.getgrgid(hub.grains.GRAINS.gid).gr_name
    except KeyError:
        hub.grains.GRAINS.groupname = None



import shutil
import socket
from typing import List

# Possible value for h_errno defined in netdb.h
HOST_NOT_FOUND = 1
NO_DATA = 4


async def _get_fqdns(hub, protocol: int) -> List[str]:
    try:
        result = socket.getaddrinfo(
            host=hub.grains.GRAINS.fqdn,
            port=None,
            family=protocol,
            proto=socket.IPPROTO_IP,
            flags=socket.AI_NUMERICSERV | socket.AI_ADDRCONFIG | socket.AI_PASSIVE,
        )
        return sorted({item[4][0] for item in result})
    except socket.gaierror as e:
        hub.log.debug(e)
    return []


async def load_localhost(hub):
    hub.grains.GRAINS.localhost = socket.gethostname()


async def load_fqdn(hub):
    await hub.grains.init.wait_for("localhost")
    # try socket.getaddrinfo to get fqdn
    if "fqdn" not in hub.grains.GRAINS:
        try:
            addrinfo = socket.getaddrinfo(
                hub.grains.GRAINS.localhost,
                0,
                socket.AF_UNSPEC,
                socket.SOCK_STREAM,
                socket.SOL_TCP,
                socket.AI_CANONNAME,
            )
            for info in addrinfo:
                # info struct [family, socktype, proto, canonname, sockaddr]
                if len(info) >= 4:
                    hub.grains.GRAINS.fqdn = info[3]
        except socket.gaierror:
            hub.grains.GRAINS.fqdn = socket.getfqdn() or "localhost"


async def load_computer_name(hub):
    hostname = shutil.which("hostname")
    if hostname and not await hub.grains.init.wait_for("computer_name"):
        hub.grains.GRAINS.computer_name = (
            await hub.exec.cmd.run(hostname)
        ).stdout.strip()
    elif await hub.grains.init.wait_for("localhost"):
        hub.grains.GRAINS.computer_name = hub.grains.GRAINS.localhost


async def load_fqdns(hub):
    if await hub.grains.init.wait_for("fqdn"):
        hub.log.debug("loading fqdn based grains")
        (
            hub.grains.GRAINS.host,
            hub.grains.GRAINS.domain,
        ) = hub.grains.GRAINS.fqdn.partition(".")[::2]

        hub.grains.GRAINS.fqdn_ip4 = await _get_fqdns(hub, socket.AF_INET)
        hub.grains.GRAINS.fqdn_ip6 = await _get_fqdns(hub, socket.AF_INET6)
        hub.grains.GRAINS.fqdns = (
            hub.grains.GRAINS.fqdn_ip4 + hub.grains.GRAINS.fqdn_ip6
        )



# -*- coding: utf-8 -*-
import asyncio
import functools
import os
import shlex
from typing import Any, Dict, List

__virtualname__ = "cmd"


def __virtual__(hub):
    return os.name == "posix", "idem-posix only runs on posix systems"


async def _sanitize_env(hub, env: Dict[str, Any]) -> Dict[str, str] or None:
    if env is None:
        return
    for bad_env_key in (k for k, v in env.items() if v is None):
        hub.log.error(
            "Environment variable '%s' passed without a value. "
            "Setting value to an empty string",
            bad_env_key,
        )
        env[bad_env_key] = ""
    return env


async def _sanitize_cwd(hub, cwd: str or None) -> str:
    # salt-minion is running as. Defaults to home directory of user under which
    # the minion is running.
    if not cwd:
        cwd = os.getcwd()

        # make sure we can access the cwd
        # when run from sudo or another environment where the euid is
        # changed ~ will expand to the home of the original uid and
        # the euid might not have access to it. See issue #1844
        if not os.access(cwd, os.R_OK):
            cwd = "/"
    else:
        # Handle edge cases where numeric/other input is entered, and would be
        # yaml-ified into non-string types
        cwd = str(cwd)

    if not os.path.isabs(cwd) or not os.path.isdir(cwd):
        raise SystemError(
            f"Specified cwd '{cwd}' either not absolute or does not exist"
        )

    return cwd


async def _sanitize_cmd(hub, cmd: str or List[str]) -> str or List[str]:
    if not isinstance(cmd, list):
        cmd = cmd.split()

    # Use shlex.quote to properly escape whitespace and special characters in strings passed to shells
    if isinstance(cmd, list):
        cmd = [shlex.quote(str(x).strip()) for x in cmd]
    else:
        cmd = shlex.quote(cmd)
    return cmd


async def _sanitize_umask(hub, umask: str) -> int or None:
    if umask is None:
        return

    _umask = str(umask).lstrip("0")

    if _umask == "":
        raise SystemError("Zero umask is not allowed.")

    try:
        return int(_umask, 8)
    except ValueError:
        raise SystemError("Invalid umask: '{0}'".format(umask))


async def _sanitize_kwargs(hub, **kwargs):
    """
    Only pass through approved kwargs
    """
    new_kwargs = {}
    if "stdin_raw_newlines" in kwargs:
        new_kwargs["stdin_raw_newlines"] = kwargs["stdin_raw_newlines"]
    return new_kwargs


async def call_run(hub, ctx):
    kwargs = ctx.get_arguments()
    umask = kwargs.get("umask")
    shell = kwargs.get("shell")
    # TODO if shell = True set the cmdline shell default to the one from grains
    await hub.grains.init.wait_for("shell")
    cmd = kwargs["cmd"]

    if shell:
        if isinstance(cmd, list):
            cmd = " ".join(cmd)
    else:
        cmd = await _sanitize_cmd(hub, cmd)
    new_kwargs = {
        "cmd": cmd,
        "cwd": await _sanitize_cwd(hub, kwargs["cwd"]),
        "env": await _sanitize_env(hub, kwargs.get("env", os.environ.copy())),
        "preexec_fn": functools.partial(os.umask, await _sanitize_umask(hub, umask))
        if umask
        else None,
        "stdout": kwargs.get("stdout"),
        "stderr": kwargs.get("stderr"),
        "shell": shell,
        "timeout": kwargs.get("timeout"),
    }
    new_kwargs.update(await _sanitize_kwargs(hub, **new_kwargs))

    return await ctx.func(hub, **new_kwargs)


async def sig_run(
    hub,
    cmd: str or List[str],
    cwd: str = None,
    shell: bool = False,
    stdin: str = None,
    stdout: int = asyncio.subprocess.PIPE,
    stderr: int = asyncio.subprocess.PIPE,
    env: Dict[str, Any] = None,
    timeout: int or float = None,
    **kwargs,
) -> Dict[str, Any]:
    pass



import os
import pwd

try:
    import getpass

    HAS_PASS = hasattr(getpass, "getuser") and hasattr(pwd, "getpwnam")
except ImportError:
    HAS_PASS = False


async def load_console_user(hub):
    if HAS_PASS:
        hub.grains.GRAINS.console_username = getpass.getuser()
        hub.grains.GRAINS.console_user = pwd.getpwnam(
            hub.grains.GRAINS.console_username
        ).pw_uid
    else:
        console = "/dev/console"
        if os.path.exists(console):
            # returns the 'st_uid' stat from the /dev/console file.
            uid = os.stat(console)[4]
            hub.grains.GRAINS.console_user = uid
        else:
            hub.grains.GRAINS.console_user = 0
        hub.grains.GRAINS.console_username = pwd.getpwuid(
            hub.grains.GRAINS.console_user
        ).pw_name



import dns.resolver

import os
import ipaddress
from typing import Any, Dict, List


async def _cidr_to_ipv4_netmask(cidr_bits: str or int) -> str:
    """
    Returns an IPv4 netmask
    """
    try:
        cidr_bits = int(cidr_bits)
        if not 1 <= cidr_bits <= 32:
            return ""
    except ValueError:
        return ""

    netmask = ""
    for idx in range(4):
        if idx:
            netmask += "."
        if cidr_bits >= 8:
            netmask += "255"
            cidr_bits -= 8
        else:
            netmask += "{0:d}".format(256 - (2 ** (8 - cidr_bits)))
            cidr_bits = 0
    return netmask


async def _ipv4_to_bits(ipaddr: str) -> str:
    """
    Accepts an IPv4 dotted quad and returns a string representing its binary
    counterpart
    """
    return "".join([bin(int(x))[2:].rjust(8, "0") for x in ipaddr.split(".")])


async def _natural_ipv4_netmask(ip_addr: str, fmt: str = "prefixlen") -> str:
    """
    Returns the "natural" mask of an IPv4 address
    """
    bits = await _ipv4_to_bits(ip_addr)

    if bits.startswith("11"):
        mask = "24"
    elif bits.startswith("1"):
        mask = "16"
    else:
        mask = "8"

    if fmt == "netmask":
        return await _cidr_to_ipv4_netmask(mask)
    else:
        return "/" + mask


async def _is_ipv4_address(ip: str) -> bool:
    try:
        ipaddress.IPv4Address(ip)
        return True
    except ipaddress.AddressValueError:
        return False


async def _is_ipv6_address(ip: str) -> bool:
    try:
        ipaddress.IPv6Address(ip)
        return True
    except ipaddress.AddressValueError:
        return False


async def _parse_options(options: Dict[str, Any]) -> Dict[str, Any]:
    ret = {}
    for key, value in options.items():
        if value and not key.startswith("_"):
            if key == "keyalgorithm":
                ret[key] = str(value)
            elif key not in ("domain", "nameservers", "search", "read_resolv_conf"):
                ret[key] = value
    return ret


async def _get_sortlist(hub, ip_addrs: List[str]) -> List[str]:
    ret = []
    for ip_raw in ip_addrs:
        try:
            ip_net = ipaddress.ip_network(ip_raw)
        except ValueError as exc:
            hub.log.error(exc)
        else:
            if "/" not in ip_raw:
                # No netmask has been provided, guess
                # the "natural" one
                if ip_net.version == 4:
                    ip_addr = str(ip_net.network_address)
                    # pylint: disable=protected-access
                    mask = await _natural_ipv4_netmask(ip_addr)
                    ip_net = ipaddress.ip_network(f"{ip_addr}{mask}", strict=False)
                if ip_net.version == 6:
                    # TODO
                    pass

            if ip_net not in ret:
                ret.append(str(ip_net))
    return ret


async def load_dns(hub):
    """
    Parse the resolver configuration file
    """
    if await hub.grains.init.wait_for("dns"):
        return
    resolv = dns.resolver.Resolver(configure=False)

    # Read known resolv.conf files
    for conf in ("/etc/resolv.conf", "/var/etc/resolv.conf"):
        if os.path.isfile(conf):
            resolv.read_resolv_conf(conf)

    hub.grains.GRAINS.dns.nameservers = resolv.nameservers
    hub.grains.GRAINS.dns.ip4_nameservers = [
        ip for ip in resolv.nameservers if await _is_ipv4_address(ip)
    ]

    hub.grains.GRAINS.dns.ip6_nameservers = [
        ip for ip in resolv.nameservers if await _is_ipv6_address(ip)
    ]

    hub.grains.GRAINS.dns.sortlist = await _get_sortlist(hub, resolv.nameservers)
    hub.grains.GRAINS.dns.domain = str(resolv.domain)
    hub.grains.GRAINS.dns.search = [str(x) for x in resolv.search]
    hub.grains.GRAINS.dns.options = await _parse_options(resolv.__dict__)



import dateutil.tz
import locale
import datetime
import sys


async def load_info(hub):
    """
    Provides
        defaultlanguage
        defaultencoding
    """
    try:
        (
            hub.grains.GRAINS.locale_info.defaultlanguage,
            hub.grains.GRAINS.locale_info.defaultencoding,
        ) = locale.getdefaultlocale()
    except Exception:  # pylint: disable=broad-except
        # locale.getdefaultlocale can ValueError!! Catch anything else it
        # might do, per #2205
        hub.grains.GRAINS.locale_info.defaultlanguage = "unknown"
        hub.grains.GRAINS.locale_info.defaultencoding = "unknown"
    hub.grains.GRAINS.locale_info.detectedencoding = sys.getdefaultencoding() or "ascii"


async def load_timezone(hub):
    hub.grains.GRAINS.locale_info.timezone = datetime.datetime.now(
        dateutil.tz.tzlocal()
    ).tzname()



# -*- coding: utf-8 -*-
import asyncio
import dict_tools
import os
from typing import Any, Dict, List

__virtualname__ = "cmd"


def __virtual__(hub):
    return os.name == "posix", "idem-posix only runs on posix systems"


async def run(
    hub,
    cmd: str or List[str],
    cwd: str = None,
    shell: bool = False,
    stdin: str = None,
    stdout: int = asyncio.subprocess.PIPE,
    stderr: int = asyncio.subprocess.PIPE,
    env: Dict[str, Any] = None,
    timeout: int or float = None,
    **kwargs,
) -> Dict[str, Any]:
    """
    Execute the passed command and return the output as a string

    :param cmd: The command to run. ex: ``ls -lart /home``

    :param cwd: The directory from which to execute the command. Defaults
        to the home directory of the user specified by ``runas`` (or the user
        under which Salt is running if ``runas`` is not specified).

    :param stdin: A string of standard input can be specified for the
        command to be run using the ``stdin`` parameter. This can be useful in
        cases where sensitive information must be read from standard input.

    :param shell: If ``False``, let python handle the positional
        arguments. Set to ``True`` to use shell features, such as pipes or
        redirection.

    :param stdout:

    :param stderr:

    :param env: Environment variables to be set prior to execution.

        .. note::
            When passing environment variables on the CLI, they should be
            passed as the string representation of a dictionary.

            .. code-block:: bash

                idem exec cmd.run 'some command' env='{"FOO": "bar"}'
    :param umask: The umask (in octal) to use when running the command.

    :param timeout: A timeout in seconds for the executed process to return.

    CLI Example:

    .. code-block:: bash

        idem exec cmd.run "command" cwd=/home
    """
    await hub.grains.init.wait_for("shell")
    ret = dict_tools.data.NamespaceDict()

    # Run the command
    if shell:
        proc = await asyncio.create_subprocess_shell(
            cmd, cwd=cwd, stdout=stdout, stderr=stderr, env=env, **kwargs
        )
    else:
        proc = await asyncio.create_subprocess_exec(
            *cmd, cwd=cwd, stdout=stdout, stderr=stderr, env=env, **kwargs
        )
    ret.pid = proc.pid

    # This is where the magic happens
    out, err = await asyncio.wait_for(proc.communicate(input=stdin), timeout=timeout)

    ret.stdout = (out or b"").decode()
    ret.stderr = (err or b"").decode()
    ret.retcode = await asyncio.wait_for(proc.wait(), timeout=timeout)
    return ret



import aiofiles
import os
import shutil


async def load_systemd(hub):
    cmd = shutil.which("systemctl")
    if cmd and shutil.which("localectl"):
        hub.log.debug("Adding systemd grains")
        ret = await hub.exec.cmd.run([cmd, "--version"])
        if ret.retcode:
            hub.log.error(ret.stderr)
        else:
            systemd_info = ret.stdout.strip().splitlines()
            hub.grains.GRAINS.systemd.version = systemd_info[0].split()[1]
            hub.grains.GRAINS.systemd.features = systemd_info[1]


async def _load_cgroup(hub) -> str:
    path = "/proc/1/cgroup"
    if os.path.exists(path):
        async with aiofiles.open(path) as fh_:
            data = await fh_.read()
            if "docker" in data:
                return "docker"
            elif "name=" in data:
                name = data.split("name=")[1]
                return name.split(":")[0]


def _load_systemd(hub) -> str:
    system = "/run/systemd/system"
    if os.path.exists(system) and os.stat(system):
        return "systemd"


async def _load_bin(hub, init_bin: str) -> str:
    supported_inits = (b"upstart", b"sysvinit", b"systemd")
    edge_len = max(len(x) for x in supported_inits) - 1
    try:
        async with aiofiles.open(init_bin, "rb") as fp_:
            edge = b""
            buf = (await fp_.read(hub.OPT.grains.file_buffer_size)).lower()
            while buf:
                if isinstance(buf, str):
                    # This makes testing easier
                    buf = buf.encode()
                buf = edge + buf
                for item in supported_inits:
                    if item in buf:
                        return item.decode("utf-8")
                edge = buf[-edge_len:]
                buf = (await fp_.read(hub.OPT.grains.file_buffer_size)).lower()
    except (IOError, OSError) as exc:
        hub.log.error(f"Unable to read from init_bin ({init_bin}): {exc}")


async def _load_cmdline(hub) -> str:
    cmdline = "/proc/1/cmdline"

    if not os.path.exists(cmdline):
        return ""

    async with aiofiles.open(cmdline) as fhr:
        init_cmdline = (await fhr.read()).replace("\x00", " ").strip().split()
        if not init_cmdline:
            # Emtpy init_cmdline
            hub.log.warning("Unable to fetch data from /proc/1/cmdline")
            return ""

    if "runit" in init_cmdline:
        return "runit"
    elif "/sbin/my_init" in init_cmdline:
        # Phusion Base docker container use runit for srv mgmt, but
        # my_init as pid1
        return "runit"
    else:
        for init in (
            "supervisord",
            # https://github.com/Yelp/dumb-init
            "dumb-init",
            # https://github.com/krallin/tini
            "tini",
        ):
            if shutil.which(init) in init_cmdline:
                return init

    init_bin = shutil.which(init_cmdline[0])
    if init_bin.endswith("bin/init"):
        return await _load_bin(hub, init_bin)
    else:
        hub.log.debug(
            f"Could not determine init system from command line: {init_cmdline}"
        )


async def load_init(hub):
    # Add init grain
    # Default to the cgroup
    hub.log.debug("Adding init grain")
    hub.grains.GRAINS.init = (
        await _load_cgroup(hub)
        or _load_systemd(hub)
        or await _load_cmdline(hub)
        or "unknown"
    )



import copy
import os
from typing import Any, Dict, List

__virtualname__ = "cmd"


def __virtual__(hub):
    return os.name == "posix", "idem-posix only runs on posix systems"


async def run(
    hub,
    ctx,
    name: str or List[str],
    cwd: str = None,
    shell: bool = False,
    env: Dict[str, Any] = None,
    umask: str = None,
    timeout: int or float = None,
    ignore_timeout: bool = False,
    stateful: bool = False,
    **kwargs,
) -> Dict[str, Any]:
    """
    Run a command if certain circumstances are met.  Use ``cmd.wait`` if you
    want to use the ``watch`` requisite.

    name
        The command to execute, remember that the command will execute with the
        path and permissions of the salt-minion.

    cwd
        The current working directory to execute the command in, defaults to
        /root

    shell
        Run the command in the shell

    stateful
        The command being executed is expected to return data about executing
        a state. For more information, see the :ref:`stateful-argument` section.

    umask
        The umask (in octal) to use when running the command.

    timeout
        If the command has not terminated after timeout seconds, send the
        subprocess sigterm, and if sigterm is ignored, follow up with sigkill

    ignore_timeout
        Ignore the timeout of commands, which is useful for running nohup
        processes.
    """
    ret = {"name": name, "changes": {}, "result": False, "comment": ""}
    test_name = None
    if isinstance(stateful, list) and "test_name" in stateful[0]:
        test_name = stateful[0]["test_name"]
    if ctx["test"] and test_name:
        name = test_name

    # Need the check for None here, if env is not provided then it falls back
    # to None and it is assumed that the environment is not being overridden.
    if env is not None and not isinstance(env, (list, dict)):
        ret["comment"] = "Invalidly-formatted 'env' parameter. See " "documentation."
        return ret

    cmd_kwargs = copy.deepcopy(kwargs)
    cmd_kwargs.update(
        {"cwd": cwd, "shell": shell, "env": env, "umask": umask,}
    )

    if ctx["test"] and not test_name:
        ret["result"] = None
        ret["comment"] = 'Command "{0}" would have been executed'.format(name)
        return ret

    if cwd and not os.path.isdir(cwd):
        ret["comment"] = f'Desired working directory "{cwd}" ' "is not available"
        return ret

    # Wow, we passed the test, run this sucker!
    try:
        cmd_all = await hub.exec.cmd.run(
            cmd=name, timeout=timeout, python_shell=True, **cmd_kwargs
        )
    except Exception as err:  # pylint: disable=broad-except
        ret["comment"] = str(err)
        return ret

    ret["changes"] = cmd_all
    ret["result"] = not bool(cmd_all.retcode)
    ret["comment"] = f'Command "{name}" run'

    # Ignore timeout errors if asked (for nohups) and treat cmd as a success
    if ignore_timeout:
        trigger = "Timed out after"
        if ret["changes"].get("retcode") == 1 and trigger in ret["changes"].get(
            "stdout"
        ):
            ret["changes"]["retcode"] = 0
            ret["result"] = True

    if ctx["test"] and cmd_all["retcode"] == 0 and ret["changes"]:
        ret["result"] = None

    return ret



CLI_CONFIG = {}
CONFIG = {
    "file_buffer_size": {
        "dyne": "grains",
        "default": 262144,
        "help": "chunk size for transported files",
        "type": int,
    }
}
SUBS = {}
DYNE = {
    "grains": ["grains"],
    "exec": ["exec"],
    "states": ["states"],
}



import os


async def load_uname(hub):
    (
        hub.grains.GRAINS.kernel,
        hub.grains.GRAINS.nodename,
        hub.grains.GRAINS.kernelrelease,
        hub.grains.GRAINS.kernelversion,
        _,
    ) = os.uname()



import json
import os
import shlex
from typing import Any, Dict

__virtualname__ = "cmd"


def __virtual__(hub):
    return os.name == "posix", "idem-posix only runs on posix systems"


def _is_true(hub, val: str) -> bool:
    if val and str(val).lower() in ("true", "yes", "1"):
        return True
    elif str(val).lower() in ("false", "no", "0"):
        return False
    raise ValueError(f"Failed parsing boolean value: {val}")


def _failout(hub, state: Dict[str, Any], msg: str) -> Dict[str, Any]:
    state["comment"] = msg
    state["result"] = False
    return state


def _reinterpreted_state(hub, state: Dict[str, Any]) -> Dict[str, Any]:
    """
    Re-interpret the state returned by salt.state.run using our protocol.
    """
    ret = state["changes"]
    state["changes"] = {}
    state["comment"] = ""

    out = ret.get("stdout")
    if not out:
        if ret.get("stderr"):
            state["comment"] = ret["stderr"]
        return state

    is_json = False
    try:
        data = json.loads(out)
        if not isinstance(data, dict):
            return _failout(
                hub, state, "script JSON output must be a JSON object (e.g., {})!"
            )
        is_json = True
    except ValueError:
        idx = out.rstrip().rfind("\n")
        if idx != -1:
            out = out[idx + 1 :]
        data = {}
        try:
            for item in shlex.split(out):
                key, val = item.split("=")
                data[key] = val
        except ValueError:
            state = _failout(
                state,
                "Failed parsing script output! "
                "Stdout must be JSON or a line of name=value pairs.",
            )
            state["changes"].update(ret)
            return state

    changed = _is_true(hub, data.get("changed", "no"))

    if "comment" in data:
        state["comment"] = data["comment"]
        del data["comment"]

    if changed:
        for key in ret:
            data.setdefault(key, ret[key])

        # if stdout is the state output in JSON, don't show it.
        # otherwise it contains the one line name=value pairs, strip it.
        data["stdout"] = "" if is_json else data.get("stdout", "")[:idx]
        state["changes"] = data

    # FIXME: if it's not changed but there's stdout and/or stderr then those
    #       won't be shown as the function output. (though, they will be shown
    #       inside INFO logs).
    return state


def post_run(hub, ctx):
    kwargs = ctx.get_arguments()
    if kwargs["stateful"]:
        return _reinterpreted_state(hub, ctx.ret)
    else:
        return ctx.ret



import aiofiles
import os


async def load_machine_id(hub):
    """
    Provide the machine-id for machine/virtualization combination
    """
    # Provides:
    #   machine-id
    locations = ["/etc/machine-id", "/var/lib/dbus/machine-id"]
    existing_locations = [loc for loc in locations if os.path.exists(loc)]
    if existing_locations:
        async with aiofiles.open(existing_locations[0]) as machineid:
            hub.grains.GRAINS.machine_id = (await machineid.read()).strip()
